part of masamune.path;

/// Abstract class of the path model.
abstract class Model implements IPath {
  /// The abstract class of the path model.
  Model() {
    this._path = Path(this.runtimeType.hashCode.toString());
  }

  /// Destroys the object.
  ///
  /// Destroyed objects are not allowed.
  @mustCallSuper
  void dispose() {
    if (this.isDisposed || !this.isDisposable) return;
    this._isDisposed = true;
    if (this._onDispose != null) this._onDispose();
    PathMap.remove(this);
    PathListener._updateNotify(this);
    Log.ast("Disposed object: %s", [this.path]);
  }

  /// Debug output.
  ///
  /// The debug output differs depending on the object.
  ///
  /// [value]: The object to enter.
  @override
  T debug<T extends IDebuggable>([Object value]) {
    if (value != null)
      Log.msg(value);
    else
      Log.msg("PathModel[%s][%s]\n%s",
          [this.runtimeType, this.path, this.toString()]);
    return this as T;
  }

  /// Get the path extension.
  final String extension = Const.empty;

  /// Get group number.
  final int group = 0;

  /// Get the path.
  String get path => this._path.url;

  /// Get ID.
  ///
  /// Usually the string of the child of the path.
  String get id {
    if (isEmpty(this._path?.file)) return Const.empty;
    return this._path.file;
  }

  /// True if the object can be destroyed.
  final bool isDisposable = true;

  /// True if the object has been destroyed.
  @override
  bool get isDisposed => this._isDisposed;
  bool _isDisposed = false;

  /// True if monitoring is possible.
  final bool isObservable = false;

  /// True if the object is temporary data.
  final bool isTemporary = false;

  /// Link with another path.
  ///
  /// You can access the same object even if you access the linked path.
  ///
  /// If the path of the actual object exists, it takes precedence.
  ///
  /// If the link destination is a link, it will be overwritten.
  ///
  /// [path]: Link destination path.
  void linkTo(String path) => PathMap.link(this.path, path);

  /// Get updated time (ms).
  int get updatedTime => this._updatedTime;
  int _updatedTime = 0;

  /// Notify object update.
  ///
  /// Notifications spread to related objects.
  ///
  /// However, updatedTime is not notified to newer objects.
  ///
  /// [updatedTime]: Updated time (ms). If it is less than 0, it will be obtained automatically from the application.
  @override
  void notifyUpdate([int updatedTime = -1]) {
    if (updatedTime < 0)
      updatedTime = DateTime.now().frameMillisecondsSinceEpoch;
    if (updatedTime <= this._updatedTime) return;
    this._updatedTime = updatedTime;
    PathListener._updateNotify(this);
  }

  /// Callback event when application pauses.
  ///
  /// [pause]: True when application is paused, False when returning from pause.
  @override
  void onApplicationPause(bool pause) {}

  /// Callback event when application quit.
  @override
  void onApplicationQuit() {}

  /// Callback executed at Dispose.
  ///
  /// [callback]: Callback.
  T onDispose<T extends IDisposable>(void callback(T disposable)) {
    if (callback != null) this._onDispose = () => callback(this as T);
    return this as T;
  }

  void Function() _onDispose;

  /// Get the protocol of the path.
  @override
  final int order = 0;

  /// Get the protocol of the path.
  @override
  final String protocol = Protocol.model;

  /// Get the path structure.
  Path get rawPath => this._path;
  Path _path;

  /// Register the temporary data in the path map.
  ///
  /// Not registered if the protocol is temporary.
  void registerUntemporary() {}
}
